import time

import jwt
from fastapi import Depends, APIRouter, Request, HTTPException
from oic import rndstr
from oic.exception import AccessDenied
from oic.oic import Client
from oic.oic.message import RegistrationResponse, AuthorizationResponse
from oic.utils.authn.client import CLIENT_AUTHN_METHOD
from starlette import status

from model.login_oic import LoginRequest, LoginTokenResponse, LoginResponse
from model.wapi.cntl import APIToken
from model.wapi.cntl import Mgr
from model.settings import settings
from . import get_conn
from util.auth import check_auth

router = APIRouter(
    prefix="/api",
    tags=["login_oic"]
)
js_mods = ['login_oic']


def get_client():
    oic_client = Client(client_authn_method=CLIENT_AUTHN_METHOD)
    oic_client.provider_config(settings.oic_endpoint)
    info = {
        'client_id': settings.oic_client_id,
        'client_secret': settings.oic_client_secret,
    }
    client_reg = RegistrationResponse(**info)
    oic_client.store_registration_info(client_reg)
    return oic_client


def get_authorization_response(client, request: Request):
    response = dict(request.query_params)  # must be dict, because authorizationResponse does not take type QueryParams
    authorization_response = client.parse_response(AuthorizationResponse, info=response, sformat="dict")
    return authorization_response


@router.post('/login')
def api_login(login_request: LoginRequest, request: Request) -> LoginResponse:
    oic_client = get_client()
    origin = str(request.base_url)

    if settings.debug:
        redirect_uri = origin.rstrip('/') + settings.oic_redirect_path
        if 'localhost' not in origin:
            redirect_uri = redirect_uri.replace('http://', 'https://')
    else:
        redirect_uri = settings.oic_redirect_host + settings.oic_redirect_path

    nonce = rndstr()
    state = jwt.encode({
        'stay_logged_in': login_request.stay_logged_in,
        'time': str(int(time.time())),
        'redirect_uri': redirect_uri,
        'nonce': nonce
    }, settings.jwt_secret, algorithm=settings.jwt_algorithm)

    args = {
        'client_id': oic_client.client_id,
        'response_type': 'code',
        'scope': ['openid', 'profile', 'email'],
        'nonce': nonce,
        'redirect_uri': redirect_uri,
        'state': state,
        'acr_values': 'phrh',
        'prompt': ['select_account']
    }

    if not settings.require_2fa:
        del args['acr_values']

    auth_req = oic_client.construct_AuthorizationRequest(request_args=args)
    login_url = auth_req.request(oic_client.authorization_endpoint)
    return LoginResponse(login=login_url, state=state)


@router.get('/oic_create_token')
def oic_create_token(request: Request, conn=Depends(get_conn)) -> LoginTokenResponse:
    params = {}
    for k in request.query_params.keys():
        params[k] = request.query_params[k]

    client = get_client()
    authorization_response = client.parse_response(
        AuthorizationResponse,
        info=params,
        sformat='dict')

    raw_state: str = authorization_response['state']
    state: dict[str, str] = jwt.decode(raw_state, settings.jwt_secret, algorithms=settings.jwt_algorithm)

    args = {
        "code": authorization_response['code'],
        "redirect_uri": state['redirect_uri'],
    }

    try:
        client.do_access_token_request(state=raw_state, request_args=args, authn_method='client_secret_basic')

        userinfo = client.do_user_info_request(state=raw_state)
        user = user_from_userinfo(userinfo)

        db_manager = Mgr(login_name=user.login_name)
        login_token = db_manager.create_session_token(conn, state['stay_logged_in'])  # returns API-Token

        return LoginTokenResponse(token=login_token)
    except AccessDenied as e:
        if e.args[0] == 'invalid_token':
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail='invalid_token'
            )
        else:
            raise e


@router.get('/me')
def me_debug(token: APIToken = Depends(check_auth),
             conn=Depends(get_conn)) -> Mgr:
    mgr = Mgr.check_token(conn, token)
    mgr.get_permissions(conn)
    mgr.get_bcds(conn)
    return mgr


def user_from_userinfo(userinfo):
    login_name = userinfo.get('preferred_username', userinfo['email'].rsplit('@', 1)[0])
    user = Mgr(login_name=login_name, email=userinfo['email'])
    if user is not None:
        return user
